@@ -1,5 +1,5 @@
 // SPDX-License-Identifier: BSD-3-Clause
-// SPDX-FileCopyrightText: 2020-2024 The Monero Project
+// SPDX-FileCopyrightText: The Monero Project
 
 #include "MainWindow.h"
 #include "ui_MainWindow.h"
@@ -13,6 +13,7 @@
 #include "dialog/AddressCheckerIndexDialog.h"
 #include "dialog/BalanceDialog.h"
 #include "dialog/DebugInfoDialog.h"
+#include "dialog/HistoryExportDialog.h"
 #include "dialog/PasswordDialog.h"
 #include "dialog/TxBroadcastDialog.h"
 #include "dialog/TxConfAdvDialog.h"
@@ -25,7 +26,7 @@
 #include "dialog/WalletCacheDebugDialog.h"
 #include "libwalletqt/AddressBook.h"
 #include "libwalletqt/rows/CoinsInfo.h"
-#include "libwalletqt/Transfer.h"
+#include "libwalletqt/rows/Output.h"
 #include "libwalletqt/TransactionHistory.h"
 #include "model/AddressBookModel.h"
 #include "plugins/PluginRegistry.h"
@@ -229,9 +230,6 @@ void MainWindow::initWidgets() {
     // [Send]
     m_sendWidget = new SendWidget(m_wallet, this);
     ui->sendWidgetLayout->addWidget(m_sendWidget);
-    // --------------
-    m_contactsWidget = new ContactsWidget(m_wallet, this);
-    ui->contactsWidgetLayout->addWidget(m_contactsWidget);
 
     // [Receive]
     m_receiveWidget = new ReceiveWidget(m_wallet, this);
@@ -240,14 +238,25 @@ void MainWindow::initWidgets() {
         m_historyWidget->setSearchText(text);
         ui->tabWidget->setCurrentIndex(this->findTab("History"));
     });
-    connect(m_contactsWidget, &ContactsWidget::fill, [this](const QString &address, const QString &description){
-        m_sendWidget->fill(address, description, 0, true);
-    });
 
     // [Coins]
     m_coinsWidget = new CoinsWidget(m_wallet, this);
     ui->coinsWidgetLayout->addWidget(m_coinsWidget);
 
+    // [Contacts]
+    m_contactsWidget = new ContactsWidget(m_wallet, this);
+    ui->contactsWidgetLayout->addWidget(m_contactsWidget);
+    connect(m_contactsWidget, &ContactsWidget::fill, [this](const QString &address, const QString &description){
+        m_sendWidget->fill(address, description, 0, true);
+        ui->tabWidget->setCurrentIndex(this->findTab("Send"));
+    });
+
+    // [Notes]
+    ui->notes->setPlainText(m_wallet->getCacheAttribute("wallet.notes"));
+    connect(ui->notes, &QPlainTextEdit::textChanged, [this] {
+       m_wallet->setCacheAttribute("wallet.notes", ui->notes->toPlainText());
+    });
+
     // [Plugins..]
     for (auto* plugin : m_plugins) {
         if (!plugin->hasParent()) {
@@ -317,10 +326,6 @@ void MainWindow::initMenu() {
     connect(ui->actionExport_CSV, &QAction::triggered, this, &MainWindow::onExportHistoryCSV);
     connect(ui->actionImportHistoryCSV, &QAction::triggered, this, &MainWindow::onImportHistoryDescriptionsCSV);
 
-    // [Wallet] -> [Contacts]
-    connect(ui->actionExportContactsCSV, &QAction::triggered, this, &MainWindow::onExportContactsCSV);
-    connect(ui->actionImportContactsCSV, &QAction::triggered, this, &MainWindow::importContacts);
-
     // [View]
     m_tabShowHideSignalMapper = new QSignalMapper(this);
     connect(ui->actionShow_Searchbar, &QAction::toggled, this, &MainWindow::toggleSearchbar);
@@ -331,13 +336,23 @@ void MainWindow::initMenu() {
     m_tabShowHideMapper["Coins"] = new ToggleTab(ui->tabCoins, "Coins", "Coins", ui->actionShow_Coins, this);
     m_tabShowHideSignalMapper->setMapping(ui->actionShow_Coins, "Coins");
 
+    // Show/Hide Contacts
+    connect(ui->actionShow_Contacts, &QAction::triggered, m_tabShowHideSignalMapper, QOverload<>::of(&QSignalMapper::map));
+    m_tabShowHideMapper["Contacts"] = new ToggleTab(ui->tabContacts, "Contacts", "Contacts", ui->actionShow_Contacts, this);
+    m_tabShowHideSignalMapper->setMapping(ui->actionShow_Contacts, "Contacts");
+
+    // Show/Hide Notes
+    connect(ui->actionShow_Notes, &QAction::triggered, m_tabShowHideSignalMapper, QOverload<>::of(&QSignalMapper::map));
+    m_tabShowHideMapper["Notes"] = new ToggleTab(ui->tabNotes, "Notes", "Notes", ui->actionShow_Notes, this);
+    m_tabShowHideSignalMapper->setMapping(ui->actionShow_Notes, "Notes");
+
     // Show/Hide Plugins..
     for (const auto &plugin : m_plugins) {
         if (plugin->parent() != "") {
             continue;
         }
 
-        auto* pluginAction = new QAction(QString("Show %1").arg(plugin->displayName()), this);
+        auto* pluginAction = new QAction(plugin->displayName(), this);
         ui->menuView->insertAction(plugin->insertFirst() ? ui->actionPlaceholderBegin : ui->actionPlaceholderEnd, pluginAction);
         connect(pluginAction, &QAction::triggered, m_tabShowHideSignalMapper, QOverload<>::of(&QSignalMapper::map));
         m_tabShowHideMapper[plugin->displayName()] = new ToggleTab(plugin->tab(), plugin->displayName(), plugin->displayName(), pluginAction, this);
@@ -351,7 +366,9 @@ void MainWindow::initMenu() {
         const auto toggleTab = m_tabShowHideMapper.value(key);
         bool show = enabledTabs.contains(key);
 
-        toggleTab->menuAction->setText((show ? QString("Hide ") : QString("Show ")) + toggleTab->name);
+        toggleTab->menuAction->setText(toggleTab->name);
+        toggleTab->menuAction->setCheckable(true);
+        toggleTab->menuAction->setChecked(show);
         ui->tabWidget->setTabVisible(ui->tabWidget->indexOf(toggleTab->tab), show);
     }
     connect(m_tabShowHideSignalMapper, &QSignalMapper::mappedString, this, &MainWindow::menuToggleTabVisible);
@@ -520,7 +537,7 @@ void MainWindow::menuToggleTabVisible(const QString &key){
 
     conf()->set(Config::enabledTabs, enabledTabs);
     ui->tabWidget->setTabVisible(ui->tabWidget->indexOf(toggleTab->tab), show);
-    toggleTab->menuAction->setText((show ? QString("Hide ") : QString("Show ")) + toggleTab->name);
+    toggleTab->menuAction->setText(toggleTab->name);
 }
 
 void MainWindow::menuClearHistoryClicked() {
@@ -555,13 +572,12 @@ void MainWindow::onWalletOpened() {
     this->setEnabled(true);
 
     // receive page
-    m_wallet->subaddress()->refresh(m_wallet->currentSubaddressAccount());
+    m_wallet->subaddress()->refresh();
     if (m_wallet->subaddress()->count() == 1) {
         for (int i = 0; i < 10; i++) {
-            m_wallet->subaddress()->addRow(m_wallet->currentSubaddressAccount(), "");
+            m_wallet->subaddress()->addRow("");
         }
     }
-    m_wallet->subaddressModel()->setCurrentSubaddressAccount(m_wallet->currentSubaddressAccount());
 
     // history page
     m_wallet->history()->refresh();
@@ -941,8 +957,8 @@ void MainWindow::onTransactionCreated(PendingTransaction *tx, const QVector<QStr
     // TODO: also check that amounts match
     tx->refresh();
     QSet<QString> outputAddresses;
-    for (const auto &output : tx->transaction(0)->outputs()) {
-        outputAddresses.insert(WalletManager::baseAddressFromIntegratedAddress(output->address(), constants::networkType));
+    for (const auto &output : tx->transaction(0).outputs) {
+        outputAddresses.insert(WalletManager::baseAddressFromIntegratedAddress(output.address, constants::networkType));
     }
     QSet<QString> destAddresses;
     for (const auto &addr : address) {
@@ -1052,8 +1068,17 @@ void MainWindow::onTransactionCommitted(bool success, PendingTransaction *tx, co
     msgBox.exec();
     if (msgBox.clickedButton() == showDetailsButton) {
         this->showHistoryTab();
-        TransactionRow *txInfo = m_wallet->history()->transaction(txid.first());
-        auto *dialog = new TxInfoDialog(m_wallet, txInfo, this);
+
+        const auto& rows = m_wallet->history()->getRows();
+        auto itr = std::find_if(rows.begin(), rows.end(),
+                [&](const TransactionRow& ti) {
+            return ti.hash == txid.first();
+        });
+        if (itr == rows.end()) {
+            return;
+        }
+
+        auto *dialog = new TxInfoDialog(m_wallet, *itr, this);
         connect(dialog, &TxInfoDialog::resendTranscation, this, &MainWindow::onResendTransaction);
         dialog->show();
         dialog->setAttribute(Qt::WA_DeleteOnClose);
@@ -1255,6 +1280,9 @@ void MainWindow::changeEvent(QEvent* event)
         if (conf()->get(Config::lockOnMinimize).toBool()) {
             this->lockWallet();
         }
+        if (conf()->get(Config::showTrayIcon).toBool() && conf()->get(Config::minimizeToTray).toBool()) {
+            this->hide();
+        }
     } else {
         QMainWindow::changeEvent(event);
     }
@@ -1307,25 +1335,6 @@ void MainWindow::onResendTransaction(const QString &txid) {
     dialog.exec();
 }
 
-void MainWindow::importContacts() {
-    const QString targetFile = QFileDialog::getOpenFileName(this, "Import CSV file", QDir::homePath(), "CSV Files (*.csv)");
-    if(targetFile.isEmpty()) return;
-
-    auto *model = m_wallet->addressBookModel();
-    QMapIterator<QString, QString> i(model->readCSV(targetFile));
-    int inserts = 0;
-    while (i.hasNext()) {
-        i.next();
-        bool addressValid = WalletManager::addressValid(i.value(), m_wallet->nettype());
-        if(addressValid) {
-            m_wallet->addressBook()->addRow(i.value(), i.key());
-            inserts++;
-        }
-    }
-
-    Utils::showInfo(this, "Contacts imported", QString("Total contacts imported: %1").arg(inserts));
-}
-
 void MainWindow::saveGeo() {
     conf()->set(Config::geometry, QString(saveGeometry().toBase64()));
     conf()->set(Config::windowState, QString(saveState().toBase64()));
@@ -1375,6 +1384,11 @@ void MainWindow::showAddressChecker() {
         return;
     }
 
+    if (!m_wallet->isAddressTorsionFree(address)) {
+        Utils::showWarning(this, "Address is not torsion-free", "This address is not compatible with reference wallets.\n\nSending to this address MAY RESULT IN A LOSS OF FUNDS.");
+        return;
+    }
+
     SubaddressIndex index = m_wallet->subaddressIndex(address);
     if (!index.isValid()) {
         // TODO: probably mention lookahead here
@@ -1431,7 +1445,7 @@ void MainWindow::importTransaction() {
     dialog.exec();
 }
 
-void MainWindow::onDeviceError(const QString &error) {
+void MainWindow::onDeviceError(const QString &error, quint64 errorCode) {
     qCritical() << "Device error: " << error;
 
     if (m_showDeviceError) {
@@ -1441,7 +1455,13 @@ void MainWindow::onDeviceError(const QString &error) {
     m_statusBtnHwDevice->setIcon(this->hardwareDeviceUnpairedIcon());
     while (true) {
         m_showDeviceError = true;
-        auto result = QMessageBox::question(this, "Hardware device", "Lost connection to hardware device. Attempt to reconnect?");
+
+        QString prompt = "Lost connection to hardware device. Attempt to reconnect?";
+        if (errorCode == 0x5515) {
+            prompt = QString("Device must be unlocked to continue scanning. Attempt to continue?");
+        }
+
+        auto result = QMessageBox::question(this, "Hardware device", prompt);
         if (result == QMessageBox::Yes) {
             bool r = m_wallet->reconnectDevice();
             if (r) {
@@ -1663,11 +1683,7 @@ void MainWindow::onSelectedInputsChanged(const QStringList &selectedInputs) {
     ui->frame_coinControl->setVisible(numInputs > 0);
 
     if (numInputs > 0) {
-        quint64 totalAmount = 0;
-        auto coins = m_wallet->coins()->coinsFromKeyImage(selectedInputs);
-        for (const auto coin : coins) {
-            totalAmount += coin->amount();
-        }
+        quint64 totalAmount = m_wallet->coins()->sumAmounts(selectedInputs);
 
         QString text = QString("Coin control active: %1 selected outputs, %2 XMR").arg(QString::number(numInputs), WalletManager::displayAmount(totalAmount));
         ui->label_coinControl->setText(text);
@@ -1705,13 +1721,8 @@ void MainWindow::onTxPoolBacklog(const QVector<quint64> &backlog, quint64 origin
 }
 
 void MainWindow::onExportHistoryCSV() {
-    QString fn = QFileDialog::getSaveFileName(this, "Save CSV file", QDir::homePath(), "CSV (*.csv)");
-    if (fn.isEmpty())
-        return;
-    if (!fn.endsWith(".csv"))
-        fn += ".csv";
-    m_wallet->history()->writeCSV(fn);
-    Utils::showInfo(this, "CSV export", QString("Transaction history exported to %1").arg(fn));
+    HistoryExportDialog dialog{m_wallet, this};
+    dialog.exec();
 }
 
 void MainWindow::onImportHistoryDescriptionsCSV() {
@@ -1729,23 +1740,6 @@ void MainWindow::onImportHistoryDescriptionsCSV() {
     }
 }
 
-void MainWindow::onExportContactsCSV() {
-    auto *model = m_wallet->addressBookModel();
-    if (model->rowCount() <= 0){
-        Utils::showInfo(this, "Unable to export contacts", "No contacts to export");
-        return;
-    }
-
-    const QString targetDir = QFileDialog::getExistingDirectory(this, "Select CSV output directory ", QDir::homePath(), QFileDialog::ShowDirsOnly);
-    if(targetDir.isEmpty()) return;
-
-    qint64 now = QDateTime::currentMSecsSinceEpoch();
-    QString fn = QString("%1/monero-contacts_%2.csv").arg(targetDir, QString::number(now / 1000));
-    if (model->writeCSV(fn)) {
-        Utils::showInfo(this, "Contacts exported successfully", QString("Exported to: %1").arg(fn));
-    }
-}
-
 void MainWindow::onCreateDesktopEntry() {
     auto msg = Utils::xdgDesktopEntryRegister() ? "Desktop entry created" : "Desktop entry not created due to an error.";
     QMessageBox::information(this, "Desktop entry", msg);
@@ -1953,7 +1947,7 @@ void MainWindow::toggleSearchbar(bool visible) {
     int currentTab = ui->tabWidget->currentIndex();
     if (currentTab == this->findTab("History"))
         m_historyWidget->focusSearchbar();
-    else if (currentTab == this->findTab("Send"))
+    else if (currentTab == this->findTab("Contacts"))
         m_contactsWidget->focusSearchbar();
     else if (currentTab == this->findTab("Receive"))
         m_receiveWidget->focusSearchbar();
@@ -1972,4 +1966,4 @@ int MainWindow::findTab(const QString &title) {
 
 MainWindow::~MainWindow() {
     qDebug() << "~MainWindow" << QThread::currentThreadId();
-}
\ No newline at end of file
+}
